use gtk4 as gtk;
use gtk::{glib, gio, prelude::*, subclass::prelude::*};

mod imp {
    use crate::{application, totp_window};
    use super::*;
    use glib::{subclass::Signal, clone, closure_local};
    use once_cell::sync::Lazy;
    
    #[derive(Debug, Default, gtk::CompositeTemplate)]
    #[template(file = "data/login_window.blp")]
    pub struct LoginWindow {
        #[template_child]
        pub forms_overlay: TemplateChild<gtk::Overlay>,
        #[template_child]
        pub forms_wrapper: TemplateChild<gtk::Box>,
        #[template_child]
        pub error_label: TemplateChild<gtk::Label>,
        #[template_child]
        pub device_entry: TemplateChild<gtk::Entry>,
        #[template_child]
        pub email_entry: TemplateChild<gtk::Entry>,
        #[template_child]
        pub password_entry: TemplateChild<gtk::PasswordEntry>,
        #[template_child]
        pub login_button: TemplateChild<gtk::Button>,
        #[template_child]
        pub api_key_entry: TemplateChild<gtk::PasswordEntry>,
        #[template_child]
        pub api_key_button: TemplateChild<gtk::Button>,
        #[template_child]
        pub device_entry_popover: TemplateChild<gtk::Popover>,
        #[template_child]
        pub loading_overlay: TemplateChild<gtk::Box>,
    }

    #[gtk::template_callbacks]
    impl LoginWindow {
        #[template_callback]
        fn login_form_changed(&self, _: &gtk::Widget) {
            let email = &self.email_entry.get().text();
            let password = &self.password_entry.get().text();
            let device = &self.device_entry.get().text();
            self.login_button.get().set_sensitive(
                !email.is_empty() && !password.is_empty() && !device.is_empty())
        }

        fn set_error(&self, error: &str) {
            self.set_is_loading(false);
            let error_label = &self.error_label.get();
            error_label.set_text(error);
            error_label.set_visible(true);
        }

        fn set_is_loading(&self, loading: bool) {
            if loading {
                self.error_label.get().set_visible(false);
            }
            self.loading_overlay.get().set_visible(loading);
            self.forms_wrapper.get().set_sensitive(!loading);
        }

        fn handle_need_totp(&self, app: &application::Application) {
            let totp_window = totp_window::TotpWindow::new(app);
            totp_window.connect_closure(
                "totp-success", false,
                closure_local!(@weak-allow-none self as opt_this =>
                    move |_: totp_window::TotpWindow| {
                        if let Some(this) = opt_this {
                            this.obj().emit_by_name::<()>("logged-in", &[]);
                        }
                    }));
            totp_window.connect_closure(
                "totp-failed", false,
                closure_local!(@weak-allow-none self as opt_this
                    => move |_: totp_window::TotpWindow, err: &str| {
                        if let Some(this) = opt_this {
                            this.set_error(err);
                        }
                    }));
            totp_window.present();
        }

        #[template_callback]
        fn login_form_submit(&self, _: &gtk::Widget) {
            let main_loop = glib::MainContext::default();
            main_loop.spawn_local(clone!(@weak self as this => async move {
                let email = &this.email_entry.get().text();
                let password = &this.password_entry.get().text();
                let device = &this.device_entry.get().text();
                if !email.is_empty() && !password.is_empty() && !device.is_empty() {
                    this.set_is_loading(true);
                    let app: application::Application =
                    this.obj().application().unwrap().downcast().unwrap();
                    let mut ctx = app.context().borrow_mut();
                    let login_res = ctx.login(email, password, device).await;
                    drop(ctx);
                    match login_res {
                        Ok(need_mfa) if need_mfa => this.handle_need_totp(&app),
                        Ok(_) => this.obj().emit_by_name::<()>("logged-in", &[]),
                        Err(e) => this.set_error(e.to_string().as_str()),
                    }
                }
            }));
        }

        fn get_api_key_from_field(&self) -> String {
            self.api_key_entry.get().text().replace('\n', "")
        }

        #[template_callback]
        fn api_key_form_changed(&self, _: &gtk::Widget) {
            let api_key = self.get_api_key_from_field();
            self.api_key_button.get().set_sensitive(!api_key.is_empty());
        }

        #[template_callback]
        fn api_key_form_submit(&self, _: &gtk::Widget) {
            let api_key = self.get_api_key_from_field();
            if !api_key.is_empty() {
                let obj = self.obj();
                let app: application::Application =
                    obj.application().unwrap().downcast().unwrap();
                let mut ctx = app.context().borrow_mut();
                ctx.set_api_key(Some(api_key.as_str()));
                drop(ctx);
                obj.emit_by_name::<()>("logged-in", &[]);
            }
        }

        #[template_callback]
        fn device_help_clicked(&self, _: gtk::EntryIconPosition) {
            self.device_entry_popover.get().popup();
        }
    }

    #[glib::object_subclass]
    impl ObjectSubclass for LoginWindow {
        const NAME: &'static str = "LoginWindow";
        type Type = super::LoginWindow;
        type ParentType = gtk::ApplicationWindow;

        fn class_init(class: &mut Self::Class) {
            class.bind_template();
            class.bind_template_callbacks();
        }

        fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
            obj.init_template();
        }
    }

    impl ObjectImpl for LoginWindow {
        fn signals() -> &'static [Signal] {
            static SIGNALS: Lazy<Vec<Signal>> = Lazy::new(||
                vec![Signal::builder("logged-in").build()]);
            SIGNALS.as_ref()
        }
    }

    impl WidgetImpl for LoginWindow {}
    impl WindowImpl for LoginWindow {}
    impl ApplicationWindowImpl for LoginWindow {}
}

glib::wrapper! {
    pub struct LoginWindow(ObjectSubclass<imp::LoginWindow>)
        @extends gtk::Widget, gtk::Window, gtk::ApplicationWindow,
        @implements gio::ActionMap, gio::ActionGroup;
}

impl LoginWindow {
    pub fn new(app: &crate::application::Application) -> Self {
        glib::Object::builder().property("application", app).build()
    }
}
